/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** dmadesc.c - Routines to create, destroy and manage DMA buffer descriptors.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl_struct.h"
#include "dmadesc.h"

static void
destroyDMATarget(
  DmaTarget* tgt)
{
  if (tgt->bRemote) {
    os_kernel_free(tgt->variant.remote.pPagesPci);
  } else {
    os_destroy_usbd(tgt->variant.local.pUsbd);
    os_kernel_free(tgt);
  }
}

OS_STATUS
ADMXRC_CreateDMATarget(
  DeviceContext* pDevCtx,
  void* buffer,
  unsigned long length,
  DmaTarget** dmatgt)
{
  DmaTarget* t;
  ADMXRC2_STATUS status;

  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  t = (DmaTarget*) os_kernel_malloc(sizeof(DmaTarget));
  if (t == NULL) {
    os_dbgprint(0, ("ADMXRC_CreateDMATarget: failed to allocate an DmaTarget\n"));
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }
  t->use = 1;
  t->bRemote = FALSE;
  t->variant.local.pBuffer = buffer;
  t->length = length;
  status = os_create_usbd(buffer, length, &t->variant.local.pUsbd);
  if (!OS_SUCCESS(status)) {
    switch (status) {
    case OS_STATUS_NOMEMORY:
      status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
      break;
    case OS_STATUS_UNSUCCESSFUL:
      status = OS_MAKE_STATUS(ADMXRC2_FAILED);
      break;
    default:
      break;
    }
    os_dbgprint(1, ("ADMXRC_CreateDMATarget: failed to allocate USBD\n"));
    os_kernel_free(t);
    return status;
  }
  *dmatgt = t;
  
  return OS_STATUS_SUCCESS;
}

OS_STATUS
ADMXRC_CreateDMATargetRemote(
  DeviceContext* pDevCtx,
  os_uint32* pPagesPci,
  unsigned long pageLength,
  unsigned int pageBits,
  unsigned long pagesSpanned,
  unsigned long bytesSpanned,
  unsigned long initOffset,
  DmaTarget** dmatgt)
{
  DmaTarget* t;

  t = (DmaTarget*) os_kernel_malloc(sizeof(DmaTarget));
  if (t == NULL) {
    os_dbgprint(0, ("ADMXRC_CreateDMATarget: failed to allocate an DmaTarget\n"));
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }
  t->use = 1;
  t->bRemote = TRUE;
  t->length = bytesSpanned;
  t->variant.remote.initOffset = initOffset;
  t->variant.remote.pageLength = pageLength;
  t->variant.remote.pageBits = pageBits;
  t->variant.remote.pPagesPci = (os_uint32*) os_kernel_malloc(pagesSpanned * sizeof(os_uint32));
  if (NULL == t->variant.remote.pPagesPci) {
    os_kernel_free(t);
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }
  memcpy(t->variant.remote.pPagesPci, pPagesPci, pagesSpanned * sizeof(os_uint32));
  *dmatgt = t;
  
  return OS_STATUS_SUCCESS;
}

/*
** Decrements a DMA target's usage counter and destroys it if the
** usage counter drops to 0.
*/
void
ADMXRC_ReleaseDMATarget(
  DeviceContext* pDevCtx,
  DmaTarget* dmatgt)
{
  OS_SPINLOCK_FLAGS f;
  os_boolean free_it;
  unsigned long use;

  os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
  use = dmatgt->use - 1;
  free_it = (use == 0);
  dmatgt->use = use;
  os_release_spinlock(&pDevCtx->dmadesc.lock, f);

  if (free_it) {
    os_dbgprint(3, ("ADMXRC_ReleaseDMATarget: destroying DMA target 0x%p\n", (void*) dmatgt));
    destroyDMATarget(dmatgt);
  }
}

/*
** Looks up a DMA target given its handle. If the DMA target exists,
** its usage counter is incremented.
*/
DmaTarget*
ADMXRC_GetDMATarget(
  DeviceContext* pDevCtx,
  unsigned long handle)
{
  DmaTarget* dmatgt;
  OS_SPINLOCK_FLAGS f;
  int index = handle - 1;

  if (handle == 0 || handle > ADMXRC_MAX_DMADESC) {
    return NULL;
  }

  index = handle - 1;

  os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
  dmatgt = pDevCtx->dmadesc.table[index];
  if (dmatgt != NULL) {
    dmatgt->use++;
  }
  os_release_spinlock(&pDevCtx->dmadesc.lock, f);

  return dmatgt;
}

void
ADMXRC_CleanupDMATargets(
  DeviceContext* pDevCtx)
{
  DmaTarget* dmatgt;
  int i;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(2, ("ADMXRC_CleanupDMATargets: entered\n"));

  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
    dmatgt = pDevCtx->dmadesc.table[i];
    if (dmatgt != NULL) {
      os_dbgprint(3, ("ADMXRC_CleanupDMATargets: cleaning up DMA descriptor %ld (dmatgt=0x%p) use=%ld\n",
		      (unsigned long) i + 1, (void*) dmatgt, (unsigned long) dmatgt->use));
      pDevCtx->dmadesc.table[i] = NULL;
      pDevCtx->dmadesc.freestack[pDevCtx->dmadesc.top++] = i;
      os_release_spinlock(&pDevCtx->dmadesc.lock, f);
      
      /* Now safe to release it */
      ADMXRC_ReleaseDMATarget(pDevCtx, dmatgt);
    } else {
      os_release_spinlock(&pDevCtx->dmadesc.lock, f);
    }
  }
}

/*
** Handle ADMXRC_IOCTL_SETUPDMA
*/
OS_STATUS
ADMXRC_SetupDMA(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_SETUPDMA* iptr;
  ADMXRC_IOCTLS_SETUPDMA* optr;
  OS_STATUS status;
  DmaTarget* dmatgt = NULL;
  OS_SPINLOCK_FLAGS f;
  int index;

  iptr = (ADMXRC_IOCTLS_SETUPDMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_SETUPDMA*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  
  status = ADMXRC_CreateDMATarget(pDevCtx, iptr->in.buffer, iptr->in.size, &dmatgt);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
  if (pDevCtx->dmadesc.top == 0) {
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);
    status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
    goto failed;
  } else {
    index = pDevCtx->dmadesc.freestack[--(pDevCtx->dmadesc.top)];
    pDevCtx->dmadesc.table[index] = dmatgt;
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);

    optr->out.dmadesc = (unsigned long) (index + 1);
    req->status.length = sizeof(optr->out);
  }

  return status;

failed:
  if (dmatgt != NULL) {
    destroyDMATarget(dmatgt);
  }
  
  return status;
}

/*
** Handle ADMXRC_IOCTL_SETUPDMAREMOTE
*/
OS_STATUS
ADMXRC_SetupDMARemote(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_SETUPDMAREMOTE* iptr;
  ADMXRC_IOCTLS_SETUPDMAREMOTE* optr;
  OS_STATUS status;
  DmaTarget* dmatgt = NULL;
  OS_SPINLOCK_FLAGS f;
  int index;
  os_uint32* pPagesPci;
  unsigned long pageLength;
  unsigned long pageBits;
  unsigned long pagesSpanned;
  unsigned long bytesSpanned;
  unsigned long initOffset;

  iptr = (ADMXRC_IOCTLS_SETUPDMAREMOTE*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_SETUPDMAREMOTE*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pageLength = iptr->in.pageLength;
  pageBits = iptr->in.pageBits;
  if (0 == pageLength || 0 == pageBits) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  if (pageBits >= (sizeof(unsigned long) * 8)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  pagesSpanned = iptr->in.pagesSpanned;
  bytesSpanned = iptr->in.bytesSpanned;
  if (0 == pagesSpanned || 0 == bytesSpanned) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  initOffset = iptr->in.initOffset;
  if (initOffset >= pageLength) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  pPagesPci = (os_uint32*) req->param.massBuf;
  if (pagesSpanned) {
    if (pPagesPci == NULL || req->param.massSize != (sizeof(os_int32) * pagesSpanned)) {
      return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    }
  }

  status = ADMXRC_CreateDMATargetRemote(pDevCtx,
					pPagesPci,
					pageLength,
					(unsigned int)pageBits,
					pagesSpanned,
				        bytesSpanned,
					initOffset,
					&dmatgt);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
  if (pDevCtx->dmadesc.top == 0) {
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);
    status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
    goto failed;
  } else {
    index = pDevCtx->dmadesc.freestack[--(pDevCtx->dmadesc.top)];
    pDevCtx->dmadesc.table[index] = dmatgt;
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);

    optr->out.dmadesc = (unsigned long) (index + 1);
    req->status.length = sizeof(optr->out);
  }

  return status;

failed:
  if (dmatgt != NULL) {
    destroyDMATarget(dmatgt);
  }
  
  return status;
}

OS_STATUS
ADMXRC_UnsetupDMA(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  ADMXRC_IOCTLS_UNSETUPDMA* iptr;
  unsigned long handle;
  DmaTarget* dmatgt;
  OS_SPINLOCK_FLAGS f;
  int index;

  iptr = (ADMXRC_IOCTLS_UNSETUPDMA*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  
  handle = iptr->in.dmadesc;
  if (handle == 0 || handle > ADMXRC_MAX_DMADESC) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
  }

  index = handle - 1;

  os_acquire_spinlock(&pDevCtx->dmadesc.lock, &f);
  dmatgt = pDevCtx->dmadesc.table[index];
  if (dmatgt == NULL) {
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);

    return OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
  } else {
    pDevCtx->dmadesc.table[index] = NULL;
    pDevCtx->dmadesc.freestack[pDevCtx->dmadesc.top++] = index;
    os_release_spinlock(&pDevCtx->dmadesc.lock, f);

    ADMXRC_ReleaseDMATarget(pDevCtx, dmatgt);

    return OS_STATUS_SUCCESS;
  }
}

/*
** x86 PC specific at the moment - how do we convert a PFN to
** a PCI address in a platform-independent manner?
*/
OS_STATUS
ADMXRC_MapDirectMaster(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  unsigned long dmadesc;
  DmaTarget* dmatgt = NULL;
  ADMXRC_IOCTLS_MAPDIRECTMASTER* iptr;
  ADMXRC_IOCTLS_MAPDIRECTMASTER* optr;
  ADMXRC2_STATUS status = ADMXRC2_SUCCESS;
  unsigned long* pagebuff;
  unsigned long length, offset;
  unsigned long maxpage;
  unsigned long pf_offs = 0; /* makes warning go away */
  unsigned long chunk;
  unsigned long pagespan, bytespan;
  os_uint32 busAddr = 0; /* makes warning go away */
  os_boolean bSuccess;

  iptr = (ADMXRC_IOCTLS_MAPDIRECTMASTER*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)){
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  optr = (ADMXRC_IOCTLS_MAPDIRECTMASTER*) req->param.outBuf;
  if (optr == NULL || req->param.outSize != sizeof(optr->out)){
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  dmadesc = iptr->in.dmadesc;
  offset = iptr->in.offset;
  length = iptr->in.length;
  maxpage = iptr->in.maxpage;

  pagebuff = (unsigned long*) req->param.massBuf;
  if (maxpage) {
    if (pagebuff == NULL || req->param.massSize != (sizeof(unsigned long) * maxpage)) {
      status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
      goto done;
    }
  }

  dmatgt = ADMXRC_GetDMATarget(pDevCtx, dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto done;
  }

  if (dmatgt->bRemote) {
    status = OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
    goto done;
  }

  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto done;
  }

  pagespan = 0;
  bytespan = 0;
  while (length && maxpage) {
    bSuccess = os_get_usbd_bus_addr(pDevCtx->pDevObj, dmatgt->variant.local.pUsbd, offset, &busAddr, &pf_offs);
    if (!bSuccess) {
      status = OS_MAKE_STATUS(ADMXRC2_FAILED);
    }
    chunk = os_page_size - pf_offs;
    if (chunk > length) {
      chunk = length;
    }
    *pagebuff++ = (unsigned long) busAddr;
    length -= chunk;
    maxpage--;
    offset += chunk;
    bytespan += chunk;
    pagespan++;
  }

  optr->out.pagelen  = os_page_size;
  optr->out.pagebits = os_page_shift;
  optr->out.pagespan = pagespan;
  optr->out.bytespan = bytespan;
  optr->out.initoffs = os_get_usbd_page_offset(dmatgt->variant.local.pUsbd);

  req->status.length = sizeof(optr->out);

  if (os_dbglevel(9)) {
    unsigned long  i;
    unsigned long* p = (unsigned long*) req->param.massBuf;

    os_dbgprint(0, ("ADMXRC_MapDirectMaster: pf_offs = 0x%ld pagespan = 0x%ld bytespan = 0x%ld\n",
		    (unsigned long) pf_offs, (unsigned long) pagespan, (unsigned long) bytespan));
    for (i = 0; i < pagespan; i++) {
      os_dbgprint(0, ("ADMXRC_MapDirectMaster: %4ld bus_addr = 0x%8.8lx\n",
		      (unsigned long) i, (unsigned long) p[i]));
    }
  }

done:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(pDevCtx, dmatgt);
  }

  return status;
}

OS_STATUS
ADMXRC_SyncDirectMaster(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* req)
{
  unsigned long dmadesc;
  DmaTarget* dmatgt = NULL;
  ADMXRC_IOCTLS_SYNCDIRECTMASTER* iptr;
  int syncmode;
  ADMXRC2_STATUS status = OS_STATUS_SUCCESS;
  unsigned long offset, length;

  iptr = (ADMXRC_IOCTLS_SYNCDIRECTMASTER*) req->param.inBuf;
  if (iptr == NULL || req->param.inSize != sizeof(iptr->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  
  dmadesc = iptr->in.dmadesc;
  syncmode = iptr->in.syncmode;
  length = iptr->in.length;
  offset = iptr->in.offset;

  dmatgt = ADMXRC_GetDMATarget(pDevCtx, dmadesc);
  if (dmatgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto done;
  }
  if (dmatgt->bRemote) {
    status = OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
    goto done;    
  }
  if (offset >= dmatgt->length || (offset + length) > dmatgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto done;
  }

  switch (syncmode) {
  case ADMXRC2_SYNC_CPUTOFPGA:
    os_flush_io_buffer(dmatgt->variant.local.pUsbd, offset, length, FALSE);
    break;
  case ADMXRC2_SYNC_FPGATOCPU:
    os_flush_io_buffer(dmatgt->variant.local.pUsbd, offset, length, TRUE);
    break;
  default:
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto done;
  }
  
done:
  if (dmatgt != NULL) {
    ADMXRC_ReleaseDMATarget(pDevCtx, dmatgt);
  }

  return status;
}
