/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2005 Alpha Data
**
** main.c - Entry points for the generic section of the driver.
**
** These entry points are called by the platform-specific section of the
** driver.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "device_defs.h"
#include "ioctl_code.h"

/* Helper routines for DMA */
#include "dmacore.h"

/* Helper routines for DMA buffer descriptors */
#include "dmadesc.h"

/* Helper routines for getting driver parameters from system */
#include "drvparam.h"

/* Helper routines for Flash access */
#include "flash.h"

/* Get the model-specific method routines */
#include "methods.h"

/* The following header files declare the IOCTL handlers */
#include "configrw.h"
#include "dma.h"
#include "flashrw.h"
#include "info.h"
#include "intwait.h"
#include "lbrst.h"
#include "loadfpga.h"
#include "owner.h"
#include "rw.h"
#include "setclock.h"
#include "spaceconfig.h"

static OS_DEVICE_DESC devDescs[32];

/*
** Declare forward references
*/

static OS_STATUS
onStopDevice(
  OS_DEVICE_OBJECT*    pDevObj);

static void
freeResources(
  DeviceContext* pDevCtx);

static void
dmaint_dpc(
  void* context,
  void* argument);

/*
** -----------------------------------------------------------------------------
** Helper routines
** -----------------------------------------------------------------------------
*/

static OS_STATUS
allocateResources(
  DeviceContext* pDevCtx)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  int       i;

  os_dbgprint(1, ("allocateResources: entered\n"));

  /* Allocate memory for DMA descriptors */
  for (i = 0; i < pDevCtx->dma.info.numChan; i++) {
    unsigned long descSize = pDevCtx->dma.info.descSize;
    unsigned long n = descSize * (ADMXRC_MAX_PLXDESC_PER_CHANNEL + 1);
    os_uint64     descPci;
    int           j;

    /* Allocate a block of PLX descriptors for the DMA channel, with one spare for alignment purposes */
    pDevCtx->dma.channel[i].descriptors.real =
      os_kernel_malloc_dma(pDevCtx->pDevObj, n, &descPci);
    if (pDevCtx->dma.channel[i].descriptors.real == NULL) {
      os_dbgprint(0, ("*** allocateResources: failed to allocate descriptors for DMA channel %d (requested %ld bytes)\n",
		      i, n));
      goto failed;
    }
    pDevCtx->dma.channel[i].descriptors.realSize = n;
    pDevCtx->dma.channel[i].descriptors.realPci = descPci;

    /* Align the PLX descriptors to a 16-byte boundary */
    pDevCtx->dma.channel[i].descriptors.aligned.untyped = (void*)
      ((((unsigned long) pDevCtx->dma.channel[i].descriptors.real) + descSize - 1) & ~(descSize - 1));

    /* Make a lookup table of the bus-relative aligned addresses of the PLX descriptors */
    descPci += ((os_uint8*) pDevCtx->dma.channel[i].descriptors.aligned.untyped) - ((os_uint8*) pDevCtx->dma.channel[i].descriptors.real);
    for (j = 0; j < ADMXRC_MAX_PLXDESC_PER_CHANNEL; j++) {
      pDevCtx->dma.channel[i].descriptors.pci[j] =
	(os_uint32) (descPci + (j * descSize));
    }
  }

  /* Allocate scatter-gather lists */
  for (i = 0; i < pDevCtx->dma.info.numChan; i++) {
    pDevCtx->dma.channel[i].sgMap = os_sg_alloc(
      pDevCtx->pDevObj, ADMXRC_MAX_PLXDESC_PER_CHANNEL);
    if (pDevCtx->dma.channel[i].sgMap == NULL) {
      os_dbgprint(0, ("*** allocateResources: failed to allocate scatter-gather list for DMA channel %d\n", i));
      goto failed;
    }
  }

  status = os_create_dpc(pDevCtx->pDevObj, ADMXRC_FpgaIntDpc, pDevCtx, &pDevCtx->interrupts.fpgaintDpc);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(0, ("*** allocateResources: failed to create DPC for FPGA interrupt\n"));
    goto failed;
  }

  for (i = 0; i < pDevCtx->dma.info.numChan; i++) {
    status = os_create_dpc(pDevCtx->pDevObj, dmaint_dpc, pDevCtx, &pDevCtx->interrupts.dmaintDpc[i]);
    if (!OS_SUCCESS(status)) {
      os_dbgprint(0, ("*** allocateResources: failed to create DPC for DMA interrupt\n"));
      goto failed;
    }
  }

  pDevCtx->dmadesc.table = (DmaTarget**) os_kernel_malloc(sizeof(DmaTarget*) * ADMXRC_MAX_DMADESC);
  if (pDevCtx->dmadesc.table == NULL) {
    os_dbgprint(0, ("*** allocateResources: failed to allocate DMA target mapping table\n"));
    goto failed;
  }
  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    pDevCtx->dmadesc.table[i] = NULL;
  }

  pDevCtx->dmadesc.freestack = (int*) os_kernel_malloc(sizeof(int) * ADMXRC_MAX_DMADESC);
  if (pDevCtx->dmadesc.freestack == NULL) {
    os_dbgprint(0, ("*** allocateResources: failed to allocate DMA descriptor free stack\n"));
    goto failed;
  }
  for (i = 0; i < ADMXRC_MAX_DMADESC; i++) {
    pDevCtx->dmadesc.freestack[i] = ADMXRC_MAX_DMADESC - 1 - i;
  }

  status = os_create_lcache(sizeof(DmaNode), 0, &pDevCtx->dma.ctlPool);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  status = os_create_lcache(sizeof(DmaNodeLoadFpga), 0, &pDevCtx->fpga.ctlPool);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  status = os_create_lcache(sizeof(IntWaitNode), 0, &pDevCtx->intwait.ctlPool);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_dbgprint(1, ("allocateResources: successful\n"));

  return status;

 failed:
  os_dbgprint(0, ("*** allocateResources: failed\n"));
  freeResources(pDevCtx);
  return status;
}

static void
freeResources(
  DeviceContext* pDevCtx)
{
  int i;

  os_dbgprint(1, ("freeResources: entered\n"));

  if (pDevCtx->intwait.ctlPool != NULL) {
    os_destroy_lcache(pDevCtx->intwait.ctlPool);
    pDevCtx->intwait.ctlPool = NULL;
  }

  if (pDevCtx->fpga.ctlPool != NULL) {
    os_destroy_lcache(pDevCtx->fpga.ctlPool);
    pDevCtx->fpga.ctlPool = NULL;
  }

  if (pDevCtx->dma.ctlPool != NULL) {
    os_destroy_lcache(pDevCtx->dma.ctlPool);
    pDevCtx->dma.ctlPool = NULL;
  }

  if (pDevCtx->dmadesc.freestack != NULL) {
    os_kernel_free(pDevCtx->dmadesc.freestack);
    pDevCtx->dmadesc.freestack = NULL;
  }

  if (pDevCtx->dmadesc.table != NULL) {
    os_kernel_free(pDevCtx->dmadesc.table);
    pDevCtx->dmadesc.table = NULL;
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    if (pDevCtx->interrupts.dmaintDpc[i] != NULL) {
      os_destroy_dpc(pDevCtx->interrupts.dmaintDpc[i]);
      pDevCtx->interrupts.dmaintDpc[i] = NULL;
    }
  }

  if (pDevCtx->interrupts.fpgaintDpc != NULL) {
    os_destroy_dpc(pDevCtx->interrupts.fpgaintDpc);
    pDevCtx->interrupts.fpgaintDpc = NULL;
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    if (pDevCtx->dma.channel[i].descriptors.real != NULL) {
      os_kernel_free_dma(
        pDevCtx->pDevObj,
        pDevCtx->dma.channel[i].descriptors.realSize,
        pDevCtx->dma.channel[i].descriptors.real,
        pDevCtx->dma.channel[i].descriptors.realPci);
      pDevCtx->dma.channel[i].descriptors.real = NULL;
    }
  }

  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    if (pDevCtx->dma.channel[i].sgMap != NULL) {
      os_sg_free(pDevCtx->dma.channel[i].sgMap);
      pDevCtx->dma.channel[i].sgMap = NULL;
    }
  }
}

static os_boolean
identifyDevice(
  DeviceContext* pDevCtx)
{
  os_boolean bIdentified = FALSE;

  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashPlxbased;
    pDevCtx->methods.flashReset = methodFlashResetPlxbased;
    pDevCtx->methods.intEnab = methodIntEnabPlxbased;
    pDevCtx->methods.isr = methodIsrPlxbased;
    pDevCtx->methods.fpga = methodFpgaPlxbased;
    pDevCtx->methods.selectMap = methodSelectMapPlxbased;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc;
    pDevCtx->methods.clockWord = methodClockWordXrc;
    pDevCtx->methods.clockProg = methodClockProgPlxbased;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc;
    pDevCtx->methods.localBusReset = methodLocalBusReset9080;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9080_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x8;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 8;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC/ADM-XRC-P\n"));
    break;
    
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc2l;
    pDevCtx->methods.flashReset = methodFlashResetPlxbased;
    pDevCtx->methods.intEnab = methodIntEnabPlxbased;
    pDevCtx->methods.isr = methodIsrPlxbased;
    pDevCtx->methods.fpga = methodFpgaPlxbased;
    pDevCtx->methods.selectMap = methodSelectMapPlxbased;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc;
    pDevCtx->methods.clockWord = methodClockWordXrc;
    pDevCtx->methods.clockProg = methodClockProgPlxbased;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc;
    pDevCtx->methods.localBusReset = methodLocalBusReset9080;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9080_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x8;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 8;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-II-Lite\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashPlxbased;
    pDevCtx->methods.flashReset = methodFlashResetPlxbased;
    pDevCtx->methods.intEnab = methodIntEnabPlxbased;
    pDevCtx->methods.isr = methodIsrPlxbased;
    pDevCtx->methods.fpga = methodFpgaPlxbased;
    pDevCtx->methods.selectMap = methodSelectMapPlxbased;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc;
    pDevCtx->methods.clockWord = methodClockWordXrc2;
    pDevCtx->methods.clockProg = methodClockProgPlxbased;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc2;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x8;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 8;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-II\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXPL_PCIX:
    if (pDevCtx->devId.pci.revision == 0) {
      /* ADM-XPL PCI-X Bridge rev. 00 */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXpl;
      pDevCtx->methods.flashReset = methodFlashResetXpl;
      pDevCtx->methods.intEnab = methodIntEnabXplrev0;
      pDevCtx->methods.isr = methodIsrXpl;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev0;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev0;
      pDevCtx->methods.doDma = methodDoDmaXplrev0;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev0;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXpl;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 0;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 4;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 8;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XPL with PCI-X Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    } else if (pDevCtx->devId.pci.revision == 1) {
      /* ADM-XPL PCI-X Bridge rev. 01 */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXpl;
      pDevCtx->methods.flashReset = methodFlashResetXpl;
      pDevCtx->methods.intEnab = methodIntEnabXplrev1;
      pDevCtx->methods.isr = methodIsrXpl;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev0;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXpl;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 4;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 8;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XPL with PCI-X Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    } else if (pDevCtx->devId.pci.revision == 2 || pDevCtx->devId.pci.revision == 3) {
      /* ADM-XPL PCI-X Bridge rev. 02 or 03 */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXpl;
      pDevCtx->methods.flashReset = methodFlashResetXpl;
      pDevCtx->methods.intEnab = methodIntEnabXplrev1;
      pDevCtx->methods.isr = methodIsrXpl;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXpl;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 4;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 8;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XPL with PCI-X Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    } else if (pDevCtx->devId.pci.revision >= 4) {
      /* ADM-XPL PCI-X Bridge rev. 04 or later */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXpl;
      pDevCtx->methods.flashReset = methodFlashResetXpl;
      pDevCtx->methods.intEnab = methodIntEnabXplrev1;
      pDevCtx->methods.isr = methodIsrXpl;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXpl;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 3;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 8;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XPL with PCI-X Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    }
    break;

  case PCI_SUBDEVICE_ADATA_ADMXPL_PCI:
    if (pDevCtx->devId.pci.revision >= 2) {
      /* ADM-XPL PCI Bridge rev. 02 or later */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXpl;
      pDevCtx->methods.flashReset = methodFlashResetXpl;
      pDevCtx->methods.intEnab = methodIntEnabXplrev1;
      pDevCtx->methods.isr = methodIsrXpl;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXpl;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 3;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 8;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XPL with PCI Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    }
    break;
    
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc2n;
    pDevCtx->methods.flashReset = methodFlashResetPlxbased;
    pDevCtx->methods.intEnab = methodIntEnabPlxbased;
    pDevCtx->methods.isr = methodIsrPlxbased;
    pDevCtx->methods.fpga = methodFpgaPlxbased;
    pDevCtx->methods.selectMap = methodSelectMapPlxbased;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc;
    pDevCtx->methods.clockWord = methodClockWordXrc2;
    pDevCtx->methods.clockProg = methodClockProgPlxbased;
    pDevCtx->methods.initCardInfo = methodInitCardInfoWrc2;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x8;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 8;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADP-WRC-II\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADPDRC2:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc2n;
    pDevCtx->methods.flashReset = methodFlashResetPlxbased;
    pDevCtx->methods.intEnab = methodIntEnabPlxbased;
    pDevCtx->methods.isr = methodIsrPlxbased;
    pDevCtx->methods.fpga = methodFpgaPlxbased;
    pDevCtx->methods.selectMap = methodSelectMapPlxbased;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc;
    pDevCtx->methods.clockWord = methodClockWordXrc2;
    pDevCtx->methods.clockProg = methodClockProgPlxbased;
    pDevCtx->methods.initCardInfo = methodInitCardInfoDrc2;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x8;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 8;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADP-DRC-II\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXP_PCI:
    if (pDevCtx->devId.pci.revision == 0) {
      /* ADM-XP PCI Bridge rev. 00 */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXprev0;
      pDevCtx->methods.flashReset = methodFlashResetXp;
      pDevCtx->methods.intEnab = methodIntEnabXp;
      pDevCtx->methods.isr = methodIsrXp;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXp;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 3;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 16;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XP with PCI Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    } else if (pDevCtx->devId.pci.revision >= 1) {
      /* ADM-XP PCI Bridge rev. 01 or later */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXprev1;
      pDevCtx->methods.flashReset = methodFlashResetXp;
      pDevCtx->methods.intEnab = methodIntEnabXp;
      pDevCtx->methods.isr = methodIsrXp;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXp;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 3;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 16;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XP with PCI Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    }
    break;
    
  case PCI_SUBDEVICE_ADATA_ADMXP_PCIX:
    if (pDevCtx->devId.pci.revision >= 2) {
      /* ADM-XP PCI-X Bridge rev. 02 or later */
      pDevCtx->methods.configMem = methodConfigMemXpl;
      pDevCtx->methods.flash = methodFlashXprev1;
      pDevCtx->methods.flashReset = methodFlashResetXp;
      pDevCtx->methods.intEnab = methodIntEnabXp;
      pDevCtx->methods.isr = methodIsrXp;
      pDevCtx->methods.fpga = methodFpgaXpl;
      pDevCtx->methods.selectMap = methodSelectMapXpl;
      pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
      pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
      pDevCtx->methods.doDma = methodDoDmaXplrev1;
      pDevCtx->methods.clockWord = methodClockWordXpl;
      pDevCtx->methods.clockProg = methodClockProgXplrev2;
      pDevCtx->methods.initCardInfo = methodInitCardInfoXp;
      pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
      pDevCtx->dma.info.numChan = 2;
      pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
      pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
      pDevCtx->miscInfo.numBar = 3;
      pDevCtx->miscInfo.selectMapOffset = 0x10;
      pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
      pDevCtx->info.flash.width = 16;
      bIdentified = TRUE;
      os_dbgprint(0, ("identifyDevice: ADM-XP with PCI-X Bridge rev %02d\n",
		      (int) pDevCtx->devId.pci.revision));
    }
    break;
    
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCI:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCIX:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXprev1;
    pDevCtx->methods.flashReset = methodFlashResetXp;
    pDevCtx->methods.intEnab = methodIntEnabXp;
    pDevCtx->methods.isr = methodIsrXp;
    pDevCtx->methods.fpga = methodFpgaXpl;
    pDevCtx->methods.selectMap = methodSelectMapXpl;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXplrev1;
    pDevCtx->methods.clockWord = methodClockWordXpl;
    pDevCtx->methods.clockProg = methodClockProgXplrev2;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXpi;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = 2;
    pDevCtx->dma.info.descSize = sizeof(XPLDMADESC);
    pDevCtx->dma.info.descAlign = sizeof(XPLDMADESC);
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x10;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADP-XPI with %s Bridge rev %02d\n",
		    pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADPXPI_PCI ? "PCI" : "PCI-X",
		    (int) pDevCtx->devId.pci.revision));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC4LS:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc4ls;
    pDevCtx->methods.flashReset = methodFlashResetXrc4ls;
    pDevCtx->methods.intEnab = methodIntEnabXrc4ls;
    pDevCtx->methods.isr = methodIsrXrc4ls;
    pDevCtx->methods.fpga = methodFpgaXrc4ls;
    pDevCtx->methods.selectMap = methodSelectMapXrc4ls;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc4ls;
    pDevCtx->methods.clockWord = methodClockWordXrc4ls;
    pDevCtx->methods.clockProg = methodClockProgXrc4ls;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc4ls;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x20000;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-4LS\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC4LX:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc4ls;
    pDevCtx->methods.flashReset = methodFlashResetXrc4ls;
    pDevCtx->methods.intEnab = methodIntEnabXrc4ls;
    pDevCtx->methods.isr = methodIsrXrc4ls;
    pDevCtx->methods.fpga = methodFpgaXrc4ls;
    pDevCtx->methods.selectMap = methodSelectMapXrc4lx;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc4ls;
    pDevCtx->methods.clockWord = methodClockWordXrc4ls;
    pDevCtx->methods.clockProg = methodClockProgXrc4ls;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc4lx;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x20000;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-4LX\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC4SX:
    pDevCtx->methods.configMem = methodConfigMemPlxbased;
    pDevCtx->methods.flash = methodFlashXrc4ls;
    pDevCtx->methods.flashReset = methodFlashResetXrc4ls;
    pDevCtx->methods.intEnab = methodIntEnabXrc4ls;
    pDevCtx->methods.isr = methodIsrXrc4ls;
    pDevCtx->methods.fpga = methodFpgaXrc4ls;
    pDevCtx->methods.selectMap = methodSelectMapXrc4sx;
    pDevCtx->methods.configSpace = methodConfigSpacePlxbased;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferPlxbased;
    pDevCtx->methods.doDma = methodDoDmaXrc4ls;
    pDevCtx->methods.clockWord = methodClockWordXrc4ls;
    pDevCtx->methods.clockProg = methodClockProgXrc4ls;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc4sx;
    pDevCtx->methods.localBusReset = methodLocalBusReset9656;
    pDevCtx->dma.info.numChan = ADMXRC_NUM_PLX9656_DMA_CHANNEL;
    pDevCtx->dma.info.descSize = sizeof(PLXDESC);
    pDevCtx->dma.info.descAlign = sizeof(PLXDESC);
    pDevCtx->miscInfo.numBar = 4;
    pDevCtx->miscInfo.selectMapOffset = 0x20000;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-4SX\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADPEXRC4FX:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrce4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrce4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrce4fx;
    pDevCtx->methods.isr = methodIsrXrce4fx;
    pDevCtx->methods.fpga = methodFpgaXrce4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrce4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrce4fx;
    pDevCtx->methods.clockWord = methodClockWordXrce4fx;
    pDevCtx->methods.clockProg = methodClockProgXrce4fx;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrce4fx;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = 2;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x18;
    pDevCtx->miscInfo.bSelectMap32Bit = FALSE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADPE-DEV\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC4FX:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrc4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrc4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrc4fx;
    pDevCtx->methods.isr = methodIsrXrc4fx;
    pDevCtx->methods.fpga = methodFpgaXrc4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrc4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrc4fx;
    pDevCtx->methods.clockWord = methodClockWordXrc4fx;
    pDevCtx->methods.clockProg = methodClockProgXrc4fx;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc4fx;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = NUM_XRC4FXBRIDGE_DMA_CHANNEL;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x20;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-4FX\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC5LX:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrc4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrc4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrc4fx;
    pDevCtx->methods.isr = methodIsrXrc4fx;
    pDevCtx->methods.fpga = methodFpgaXrc4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrc4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrc4fx;
    pDevCtx->methods.clockWord = methodClockWordXrc5lx;
    pDevCtx->methods.clockProg = methodClockProgXrc5lx;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc5lx;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = NUM_XRC4FXBRIDGE_DMA_CHANNEL;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x20;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-5LX\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC5T1:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrc4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrc4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrc4fx;
    pDevCtx->methods.isr = methodIsrXrc4fx;
    pDevCtx->methods.fpga = methodFpgaXrc4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrc4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrc4fx;
    pDevCtx->methods.clockWord = methodClockWordXrc4fx;
    pDevCtx->methods.clockProg = methodClockProgXrc4fx;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc5t1;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = NUM_XRC4FXBRIDGE_DMA_CHANNEL;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x20;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-5T1\n"));
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC5LXA:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrc4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrc4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrc4fx;
    pDevCtx->methods.isr = methodIsrXrc4fx;
    pDevCtx->methods.fpga = methodFpgaXrc4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrc4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrc4fx;
    pDevCtx->methods.clockWord = methodClockWordXrc5lxa;
    pDevCtx->methods.clockProg = methodClockProgXrc5lxa;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrc5lxa;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = NUM_XRC4FXBRIDGE_DMA_CHANNEL;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x20;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADM-XRC-5LXA\n"));

  case PCI_SUBDEVICE_ADATA_ADCPXRC4LX:
    pDevCtx->methods.configMem = methodConfigMemXpl;
    pDevCtx->methods.flash = methodFlashXrc4fx;
    pDevCtx->methods.flashReset = methodFlashResetXrc4fx;
    pDevCtx->methods.intEnab = methodIntEnabXrc4fx;
    pDevCtx->methods.isr = methodIsrXrc4fx;
    pDevCtx->methods.fpga = methodFpgaXrc4fx;
    pDevCtx->methods.selectMap = methodSelectMapXrc4fx;
    pDevCtx->methods.configSpace = methodConfigSpaceXplrev1;
    pDevCtx->methods.addDmaBuffer = methodAddDmaBufferXplrev1;
    pDevCtx->methods.doDma = methodDoDmaXrc4fx;
    pDevCtx->methods.clockWord = methodClockWordXrcp4lx;
    pDevCtx->methods.clockProg = methodClockProgXrcp4lx;
    pDevCtx->methods.initCardInfo = methodInitCardInfoXrcp4lx;
    pDevCtx->methods.localBusReset = methodLocalBusResetXpl;
    pDevCtx->dma.info.numChan = NUM_XRC4FXBRIDGE_DMA_CHANNEL;
    pDevCtx->dma.info.descAlign = 0x20;
    pDevCtx->dma.info.descSize = 0x20;
    pDevCtx->miscInfo.numBar = 3;
    pDevCtx->miscInfo.selectMapOffset = 0x20;
    pDevCtx->miscInfo.bSelectMap32Bit = TRUE;
    pDevCtx->info.flash.width = 16;
    bIdentified = TRUE;
    os_dbgprint(0, ("identifyDevice: ADCP-XRC-4LX\n"));

  default:
    break;
  }

  if (!bIdentified) {
    os_dbgprint(0, ("*** identifyDevice: failed, subVendor=0x%4.4lx subDevice=0x%4.4lx revision=0x%4.4lx\n",
		    (unsigned long) pDevCtx->devId.pci.subVendor,
		    (unsigned long) pDevCtx->devId.pci.subDevice,
		    (unsigned long) pDevCtx->devId.pci.revision));
  }

  return bIdentified;
}

static void
initDeviceContext(
  DeviceContext* pDevCtx)
{
  unsigned int i;

  /* Interrupts are initially disabled */
  pDevCtx->interrupts.enabled = FALSE;

  /* Init shadow register lock */
  os_init_spinlock(&pDevCtx->shadow.lock);
  
  /* BARs are initially not mapped */
  for (i = 0; i < 6; i++) {
    pDevCtx->bar[i] = NULL;
  }

  /*
  ** Init FPGA stuff
  */
  pDevCtx->fpga.ctlPool = NULL;
  pDevCtx->fpga.busy = FALSE;
  os_init_spinlock(&pDevCtx->fpga.lock);

  /*
  ** Init clock generator stuff
  */
  pDevCtx->clock.busy = FALSE;
  os_init_spinlock(&pDevCtx->clock.lock);

  /*
  ** Init interrupt wait stuff
  */
  pDevCtx->intwait.signalled = FALSE;
  pDevCtx->intwait.state = 0;
  pDevCtx->intwait.waiting = NULL;
  pDevCtx->intwait.eventList = NULL;
  pDevCtx->intwait.ctlPool = NULL;
  os_init_spinlock(&pDevCtx->intwait.lock);

  /*
  ** Init client lists
  */
  pDevCtx->owner.exclusive = NULL;
  pDevCtx->owner.numShared = 0;
  os_init_spinlock(&pDevCtx->owner.lock);

  /*
  ** Init DMA descriptor stuff
  */
  pDevCtx->dmadesc.table = NULL;
  pDevCtx->dmadesc.freestack = NULL;
  pDevCtx->dmadesc.top = ADMXRC_MAX_DMADESC;
  os_init_spinlock(&pDevCtx->dmadesc.lock);

  /*
  ** Init DMA channel stuff
  */
  os_init_spinlock(&pDevCtx->dma.lock);
  pDevCtx->dma.ctlPool = NULL;
  for (i = 0; i <= ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    pDevCtx->dma.queue[i].head = pDevCtx->dma.queue[i].tail = NULL;
    pDevCtx->dma.queue[i].length = 0;
  }
  for (i = 0; i < ADMXRC_NUM_DMA_CHANNEL_MAX; i++) {
    pDevCtx->dma.channel[i].descriptors.real = NULL;
    pDevCtx->dma.channel[i].sgMap = NULL;
    pDevCtx->dma.channel[i].waitsubmit = NULL;
    pDevCtx->dma.channel[i].active = NULL;
    pDevCtx->dma.channel[i].waitcollect = NULL;
  }

  /*
  ** Set all methods to the error method. These should get
  ** overridden later on with pointers to the correct methods
  ** for the device model that the device context represents.
  */
  pDevCtx->methods.configMem = (MethodConfigMem*) methodError;
  pDevCtx->methods.flash = (MethodFlash*) methodError;
  pDevCtx->methods.intEnab = (MethodIntEnab*) methodError;
  pDevCtx->methods.isr = (MethodIsr*) methodError;
  pDevCtx->methods.fpga = (MethodFpga*) methodError;
  pDevCtx->methods.selectMap = (MethodSelectMap*) methodError;
  pDevCtx->methods.configSpace = (MethodConfigSpace*) methodError;
  pDevCtx->methods.addDmaBuffer = (MethodAddDmaBuffer*) methodError;
  pDevCtx->methods.doDma = (MethodDoDma*) methodError;
  pDevCtx->methods.clockWord = (MethodClockWord*) methodError;
  pDevCtx->methods.clockProg = (MethodClockProg*) methodError;

  /*
  ** Init Flash stuff
  */
  pDevCtx->info.flash.present = FALSE;
  pDevCtx->info.flash.buffer = NULL;
}

static void*
enableInterrupts(
  void* context)
{
  DeviceContext* pDevCtx = (DeviceContext*) context;
  MethodResult   res;

  if (!pDevCtx->interrupts.enabled) {
    res = pDevCtx->methods.intEnab(pDevCtx, TRUE);
    pDevCtx->interrupts.enabled = TRUE;
  }
  return (void*) TRUE;
}

static void*
disableInterrupts(
  void* context)
{
  DeviceContext* pDevCtx = (DeviceContext*) context;
  MethodResult   res;

  if (pDevCtx->interrupts.enabled) {
    res = pDevCtx->methods.intEnab(pDevCtx, FALSE);
    pDevCtx->interrupts.enabled = FALSE;
  }
  return (void*) TRUE;
}

static void
initCard9080(
  DeviceContext* pDevCtx)
{
  os_dbgprint(1, ("initCard9080: entered\n"));

  /* Enable Direct Master access */
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMRR,   0xFF800000U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMLBAM, 0xFF000000U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMLBAI, 0x0U);
  os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMPBAM, 0x1U);

  if (admxrcDriverParams.pciReadMultiple) {
    /* Use PCI read multiple commands when a bus master */
    pDevCtx->shadow.regs.plxbased.cntrl = os_read_register_le32(&pDevCtx->hw.adapter.pci9080->CNTRL);
    pDevCtx->shadow.regs.plxbased.cntrl &= ~0xf0fU;
    pDevCtx->shadow.regs.plxbased.cntrl |= 0xc0cU;
    os_write_register_le32(&pDevCtx->hw.adapter.pci9080->CNTRL, pDevCtx->shadow.regs.plxbased.cntrl);
  }
}

static void
initCard9656(
  DeviceContext* pDevCtx)
{
  /*
  ** Currently, as far as this driver is concerned, card initialization
  ** for the PCI9656 is the same as for the PCI9080, but with a few extra
  ** bits and pieces.
  */
  initCard9080(pDevCtx);

  if (ADMXRC_9656_ERRATA_WORKAROUND) {
    os_uint32 val;

    os_dbgprint(1, ("initCard9656: using PCI9656 rev AC/AD workarounds\n"));

    /*
    ** Implement workarounds for PCI9656 rev AC and AD errata
    */
    
    /* DMPBAM register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->DMPBAM);
    val &= ~(0x1U <<  2); /* Disable Direct Master read-ahead */
    /* Do not prefetch more than 8 QWORDs for a Direct Master read */
    val &= ~(0x1U <<  3);
    val |=  (0x1U << 12);
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->DMPBAM, val);
    
    /* EROMBA register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->EROMBA);
    val &= ~(0x1U <<  4); /* Don't use local bus backoff for Direct Master transfers */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->EROMBA, val);
    
    /* LBRD0 register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD0);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 0 reads */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD0, val);
    
    /* LBRD1 register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD1);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 1 reads */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->LBRD1, val);
    
    /* LMISC2 register */
    val = os_read_register_le8(&pDevCtx->hw.adapter.pci9656->LMISC2);
    val |=  (0x1U <<  5); /* Reserve 1 location in Direct Slave write FIFO */
    os_write_register_le8(&pDevCtx->hw.adapter.pci9656->LMISC2, val);
    
    /* MARBR register */
    val = os_read_register_le32(&pDevCtx->hw.adapter.pci9656->MARBR);
    val &= ~(0x1U << 16); /* Disable local bus latency timer */
    val &= ~(0x1U << 23); /* Assert REQ# on PCI for entire DMA transfer */
    val |=  (0x1U << 24); /* Operate in PCI 2.1 mode */
    val |=  (0x1U << 26); /* Flush Direct Slave read FIFO after Direct Slave write */
    val &= ~(0x1U << 28); /* Direct slave read ahead mode disabled */
    os_write_register_le32(&pDevCtx->hw.adapter.pci9656->MARBR, val);
  }
}

static void
initCard(
  DeviceContext* pDevCtx)
{
  os_dbgprint(1, ("initCard: entered\n"));
  
  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    initCard9080(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LS:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC4SX:
    initCard9656(pDevCtx);
    break;

  case PCI_SUBDEVICE_ADATA_ADMXPL_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXPL_PCI:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCI:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCIX:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCI:
  case PCI_SUBDEVICE_ADATA_ADMXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADCPXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADPEXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5T1:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LXA:
    /* Nothing to do yet */
    break;

  default:
    os_dbgprint(0, ("*** initCard: unrecognised hardware ID: vId=0x%4.4lx dId=0x%4.4lx subVid=0x%4.4lx subDevice0x%4.4lx revision=0x%2.2lx\n",
		    (unsigned long) pDevCtx->devId.pci.vendor,
		    (unsigned long) pDevCtx->devId.pci.device,
		    (unsigned long) pDevCtx->devId.pci.subVendor,
		    (unsigned long) pDevCtx->devId.pci.subDevice,
		    (unsigned long) pDevCtx->devId.pci.revision));
    break;
  }
}

/*
** -----------------------------------------------------------------------------
** Interrupt handlers
** -----------------------------------------------------------------------------
*/

static void
dmaint_dpc(
  void* context,
  void* argument)
{
  DeviceContext* pDevCtx = (DeviceContext*) context;
  unsigned char chan = (unsigned char) (unsigned long) argument;

  ADMXRC_DmaCompleted(pDevCtx, chan);
}

static os_boolean
onInterrupt(
  void* context)
{
  DeviceContext* pDevCtx = (DeviceContext*) context;

  if (!pDevCtx->interrupts.enabled) {
    return FALSE;
  }

  return pDevCtx->methods.isr(pDevCtx);
}

/*
** -----------------------------------------------------------------------------
** OS_DRIVER_OBJECT methods
** -----------------------------------------------------------------------------
*/

static void
onDeleteDevice(
  OS_DEVICE_OBJECT*    pDevObj)
{
  DeviceContext* pDevCtx;

  os_dbgprint(1, ("onDeleteDevice: entered\n"));
  pDevCtx = os_get_device_context(pDevObj);
  if (pDevCtx != NULL) {
    freeResources(pDevCtx);  
    os_kernel_free(pDevCtx);
  }
}

static OS_STATUS
onCreateDevice(
  OS_DRIVER_OBJECT*    pDrvObj,
  OS_DEVICE_OBJECT*    pDevObj,
  int                  matchIndex,
  OS_BUSTYPE           busType,
  OS_SLOT_ID*          pSlotId,
  OS_DEVICE_ID*        pDevId)
{
  DeviceContext* pDevCtx = NULL;
  OS_STATUS      status = OS_STATUS_SUCCESS;
  os_boolean     bSuccess;

  os_dbgprint(1, ("onCreateDevice: entered\n"));

  switch (busType) {
  case OS_BUSTYPE_PCI:
    os_dbgprint(0, ("onCreateDevice: PCI device, id=%4.4lx:%4.4lx subId=%4.4lx:%4.4lx revision=%2.2lx\n",
		    (unsigned long) pDevId->pci.vendor, (unsigned long) pDevId->pci.device,
		    (unsigned long) pDevId->pci.subVendor, (unsigned long) pDevId->pci.subDevice,
		    (unsigned long) pDevId->pci.revision));
    break;

  default:
    os_dbgprint(0, ("*** onCreateDevice: don't know how to handle device of type %d\n", busType));
    status = OS_STATUS_CONFIGERROR;
    goto failed;
  }
  
  pDevCtx = (DeviceContext*) os_kernel_malloc(sizeof(DeviceContext));
  if (pDevCtx == NULL) {
    os_dbgprint(0, ("*** onCreateDevice: failed to allocate memory for device context\n"));
    status = OS_STATUS_NOMEMORY;
    goto failed;
  }

  /* Important to zero the entire device context to avoid uninitialised variables */
  memset(pDevCtx, 0, sizeof(DeviceContext));

  pDevCtx->pDrvObj = pDrvObj;
  pDevCtx->pDevObj = pDevObj;
  memcpy(&pDevCtx->devId, pDevId, sizeof(OS_DEVICE_ID));
  memcpy(&pDevCtx->slotId, pSlotId, sizeof(OS_SLOT_ID));
  
  /* Init the device context structure */
  initDeviceContext(pDevCtx);

  /* Do init according to model */
  bSuccess = identifyDevice(pDevCtx);
  if (!bSuccess) {
    status = OS_STATUS_CONFIGERROR;
    goto failed;
  }

  status = allocateResources(pDevCtx);
  if (!OS_SUCCESS(status)) {
    goto failed;
  }

  os_set_device_context(pDevObj, pDevCtx);

  os_dbgprint(1, ("onCreateDevice: successful\n"));
  return status;

 failed:
  os_dbgprint(0, ("*** onCreateDevice: failed\n"));
  pDrvObj->onDeleteDevice(pDevObj);
  return status;
}

static OS_STATUS
onStartDevice(
  OS_DEVICE_OBJECT*    pDevObj,
  OS_DEVICE_RESOURCES* pResources)
{
  unsigned long  lengths[4];
  os_boolean     map[4] = { FALSE, FALSE, FALSE, FALSE };
  DeviceContext* pDevCtx;
  unsigned int   i;
  OS_STATUS      status = OS_STATUS_SUCCESS;

  os_dbgprint(1, ("onStartDevice: entered\n"));

  pDevCtx = os_get_device_context(pDevObj);

  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    map[0] = map[2] = map[3] = TRUE;
    map[1] = FALSE;
    lengths[0] = 0x100;
    lengths[1] = 0x0;
    lengths[2] = 0x400000;
    lengths[3] = 0x400000;
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LS:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC4SX:
    map[0] = map[2] = map[3] = TRUE;
    map[1] = FALSE;
    lengths[0] = 0x200;
    lengths[1] = 0x0;
    lengths[2] = 0x400000;
    lengths[3] = 0x400000;
    break;

  case PCI_SUBDEVICE_ADATA_ADMXPL_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXPL_PCI:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCIX:
  case PCI_SUBDEVICE_ADATA_ADMXP_PCI:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCIX:
  case PCI_SUBDEVICE_ADATA_ADPXPI_PCI:
    map[0] = TRUE;
    lengths[0] = 0x100;
    if (pDevCtx->miscInfo.numBar > 3) {
      /* I/O space BAR - must not be mapped */
      map[1] = FALSE;
      lengths[1] = 0x0;
    }
    lengths[pDevCtx->miscInfo.numBar - 2] = 0x400000;
    map[pDevCtx->miscInfo.numBar - 2] = TRUE;
    lengths[pDevCtx->miscInfo.numBar - 1] = 0x400000;
    map[pDevCtx->miscInfo.numBar - 1] = TRUE;
    break;

  case PCI_SUBDEVICE_ADATA_ADPEXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADMXRC4FX:
  case PCI_SUBDEVICE_ADATA_ADCPXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC5T1:
  case PCI_SUBDEVICE_ADATA_ADMXRC5LXA:
    map[0] = TRUE;
    lengths[0] = 0x100;
    if (pDevCtx->miscInfo.numBar > 3) {
      /* I/O space BAR - must not be mapped */
      map[1] = FALSE;
      lengths[1] = 0x0;
    }
    lengths[pDevCtx->miscInfo.numBar - 2] = 0x400000;
    map[pDevCtx->miscInfo.numBar - 2] = TRUE;
    lengths[pDevCtx->miscInfo.numBar - 1] = 0x100000;
    map[pDevCtx->miscInfo.numBar - 1] = TRUE;
    break;
  }

  for (i = 0; i < pResources->raw.pci.count; i++) {
    os_dbgprint(3, ("            BAR%d io = %d physbase = 0x%8.8lx%8.8lx physlen = 0x%8.8lx0x%8.8lx\n",
		    i,
		    pResources->raw.pci.bIo[i],
		    (unsigned long) os_uint64_hi(pResources->raw.pci.base[i]),
		    (unsigned long) os_uint64_lo(pResources->raw.pci.base[i]),
		    (unsigned long) os_uint64_hi(pResources->raw.pci.size[i]),
		    (unsigned long) os_uint64_lo(pResources->raw.pci.size[i])));
  }
  os_dbgprint(3, ("            IRQ phys = %ld\n", (unsigned long) pResources->raw.pci.irqLine));
  for (i = 0; i < pResources->translated.pci.count; i++) {
    os_dbgprint(3, ("            BAR%d io = %d sysbase  = 0x%8.8lx%8.8lx syslen  = 0x%8.8lx%8.8lx\n",
		    i,
		    pResources->translated.pci.bIo[i],
		    (unsigned long) os_uint64_hi(pResources->translated.pci.base[i]),
		    (unsigned long) os_uint64_lo(pResources->translated.pci.base[i]),
		    (unsigned long) os_uint64_hi(pResources->translated.pci.size[i]),
		    (unsigned long) os_uint64_lo(pResources->translated.pci.size[i])));
  }
  os_dbgprint(3, ("            IRQ sys  = %ld\n", (unsigned long)pResources->translated.pci.irqVector));

  /* Save a copy of the pResourcess */
  memcpy(&pDevCtx->resources, pResources, sizeof(OS_DEVICE_RESOURCES));

  /*
  ** Check that the BARs have at least the minimum sizes
  */
  if (pDevCtx->resources.translated.pci.count < pDevCtx->miscInfo.numBar) {
    os_dbgprint(0, ("onStartDevice: fewer than %d BARs\n", (int) pDevCtx->miscInfo.numBar));
    status = OS_STATUS_CONFIGERROR;
    goto failed;
  }
  for (i = 0; i < pDevCtx->miscInfo.numBar; i++) {
    if (pDevCtx->resources.translated.pci.size[i] < lengths[i]) {
      os_dbgprint(0, ("onStartDevice: BAR%d is less than 0x%8.8lx%8.8lx bytes long\n",
		      i,
		      (unsigned long) os_uint64_hi(pDevCtx->resources.translated.pci.size[i]),
		      (unsigned long) os_uint64_lo(pDevCtx->resources.translated.pci.size[i])));
      status = OS_STATUS_CONFIGERROR;
      goto failed;
    }
  }

  /*
  ** Map the BARs into kernel address space
  */
  for (i = 0; i < pDevCtx->miscInfo.numBar; i++) {
    if (map[i]) {
      os_dbgprint(1, ("onStartDevice: mapping BAR%d\n", i));
      pDevCtx->bar[i] = os_map_kernel(pDevCtx->resources.translated.pci.base[i],
				      pDevCtx->resources.translated.pci.size[i]);
      if (pDevCtx->bar[i] == NULL) {
	os_dbgprint(0, ("onStartDevice: failed to map BAR%d\n", i));
	status = OS_STATUS_CONFIGERROR;
	goto failed;
      }
    }
  }

#if !defined(NO_INTERRUPTS)
  if (pDevCtx->resources.translated.pci.bIrqValid) {
    /*
    ** Attach the interrupt handler
    */
    os_dbgprint(1, ("onStartDevice: attaching interrupt handler\n"));
    status = os_attach_interrupt(pDevObj,
				 onInterrupt,
				 pDevCtx->resources.translated.pci.irqNumber,
				 pDevCtx->resources.translated.pci.irqVector,
				 DRIVER_NAME,
				 &pDevCtx->irqObj);
    if (!OS_SUCCESS(status)) {
      os_dbgprint(0, ("*** onStartDevice: failed to attach interrupt handler\n"));
      goto failed;
    }
  } else {
    os_dbgprint(0, ("*** onStartDevice: no interrupt line, not attaching interrupt handler!\n"));
    goto failed; 
  }
#endif

  pDevCtx->hw.adapter.untyped = (void*) pDevCtx->bar[0];
  pDevCtx->hw.fpga = (void*) pDevCtx->bar[pDevCtx->miscInfo.numBar - 2];
  pDevCtx->hw.space1 = (void*) pDevCtx->bar[pDevCtx->miscInfo.numBar - 1];
  pDevCtx->hw.control.untyped = pDevCtx->hw.space1;
  if (pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADPXPI_PCI ||
      pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADPXPI_PCIX) {
    /* Special case for ADP-XPI */
    pDevCtx->hw.control.byte += pDevCtx->resources.translated.pci.size[pDevCtx->miscInfo.numBar - 1];
    pDevCtx->hw.control.byte -= 0x400000;
  }

  /*
  ** Initialise the registers on the card.
  */
  initCard(pDevCtx);

  /* Identify the Flash device fitted - if any */
  identifyFlash(pDevCtx);
  if (pDevCtx->info.flash.present) {
    pDevCtx->info.flash.buffer = os_kernel_malloc(pDevCtx->info.flash.layout.eraseBlock);
    if (pDevCtx->info.flash.buffer == NULL) {
      os_dbgprint(0, ("*** onStartDevice: Failed to allocate 0x%lx bytes for Flash block buffer\n",
		      (unsigned long) pDevCtx->info.flash.layout.eraseBlock));
      status = OS_STATUS_NOMEMORY;
      goto failed;
    }
  }

  /*
  ** Obtain the physical configuration of the card, ie. how much memory etc.
  ** by reading the EEPROM (or otherwise)
  */
  if (pDevCtx->methods.initCardInfo(pDevCtx) != MethodResult_Success) {
    if (ADMXRC_IGNORE_DEVICE_IF_NO_VPD) {
      os_dbgprint(0, ("*** onStartDevice: Failed to read card VPD, treating as fatal\n"));
      status = OS_STATUS_CONFIGERROR;
      goto failed;
    } else {
      os_dbgprint(0, ("*** onStartDevice: Failed to read card VPD, continuing to allow debug\n"));
    }
  }

  /*
  ** Declare the regions that are mappable to user space
  **
  ** space 0 => FPGA space
  ** space 1 => control (CPLD) space
  ** space 2 => Bridge registers space only mapped when PCI9656RegsAsSpace driver
  **            parameter = 1
  */
  pDevCtx->userMapped.count = 0;
  for (i = 0; i < pDevCtx->info.cardInfo.numSpace; i++) {
    if (i < 2) {
      status = os_allow_mapping(pDevObj,
				pDevCtx->resources.translated.pci.base[i + pDevCtx->miscInfo.numBar - 2],
				pDevCtx->resources.translated.pci.size[i + pDevCtx->miscInfo.numBar - 2],
				(void*) (unsigned long) i, /* context value */
				&pDevCtx->userMapped.regions[i]);
      pDevCtx->userMapped.count++;
    } else if (i == 2) {
      status = os_allow_mapping(pDevObj,
				pDevCtx->resources.translated.pci.base[0],
				pDevCtx->resources.translated.pci.size[0],
				(void*) (unsigned long) i, /* context value */
				&pDevCtx->userMapped.regions[i]);
      pDevCtx->userMapped.count++;
    }
  }

  /* Enable interrupts - done with interrupts disabled to avoid races */
  os_invoke_atomic(pDevCtx->irqObj, enableInterrupts, pDevCtx);

  os_dbgprint(1, ("onStartDevice: successful, status=0x%8.8lx\n",
		  (unsigned long) status));
  return status;

 failed:
  onStopDevice(pDevObj);
  return status;
}

static OS_STATUS
onStopDevice(
  OS_DEVICE_OBJECT*    pDevObj)
{
  DeviceContext* pDevCtx;
  int       i;
  OS_STATUS status = OS_STATUS_SUCCESS;

  os_dbgprint(1, ("onStopDevice: entered\n"));

  pDevCtx = os_get_device_context(pDevObj);

  /* Free flash block buffer */
  if (pDevCtx->info.flash.buffer != NULL) {
    os_dbgprint(1, ("onStopDevice: freeing Flash block buffer at 0x%p\n",
		    (void*) pDevCtx->info.flash.buffer));
    os_kernel_free(pDevCtx->info.flash.buffer);
    pDevCtx->info.flash.buffer = NULL;
  }

  if (pDevCtx->irqObj != NULL) {
    /* Disable interrupts - done with interrupts disabled to avoid races */
    os_dbgprint(1, ("onStopDevice: disabling interrupts atomically\n"));
    os_invoke_atomic(pDevCtx->irqObj, disableInterrupts, pDevCtx);
    /* Detach the interrupt handler */
    os_dbgprint(1, ("onStopDevice: detaching interrupt handler\n"));
    os_detach_interrupt(pDevCtx->irqObj); 
    pDevCtx->irqObj = NULL;
  }

  /*
  ** Unmap regions from user space
  */
  for (i = 0; i < pDevCtx->userMapped.count; i++) {
    if (pDevCtx->userMapped.regions[i] != NULL) {
      os_dbgprint(1, ("onStopDevice: disallowing userspace mapping for region %d\n", (int) i));
      os_disallow_mapping(pDevObj, pDevCtx->userMapped.regions[i]);
      pDevCtx->userMapped.regions[i] = NULL;
    } else {
      os_dbgprint(0, ("onStopDevice: region %d is NULL\n", i));
    }
  }

  /*
  ** Unmap the BARs from kernel address space
  */
  for (i = 0; i < pDevCtx->miscInfo.numBar; i++) {
    if (pDevCtx->bar[i] != NULL) {
      os_dbgprint(1, ("onStopDevice: unmapping BAR%d\n", i));
      os_unmap_kernel(pDevCtx->bar[i],
		      pDevCtx->resources.translated.pci.base[i],
		      pDevCtx->resources.translated.pci.size[i]);
      pDevCtx->bar[i] = NULL;
    }
  }

  return status;
}

static OS_STATUS
onOpen(
  OS_CLIENT_OBJECT* pClObj)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_DEVICE_OBJECT* pDevObj;
  DeviceContext* pDevCtx;
  ClientContext* pClCtx;
  unsigned int i;

  os_dbgprint(1, ("onOpen: entered, client = 0x%p\n", pClObj));

  pDevObj = os_get_device_object(pClObj);
  pDevCtx = os_get_device_context(pDevObj);

  pClCtx = (ClientContext*) os_kernel_malloc(sizeof(ClientContext));
  if (pClCtx == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
    goto done;
  }
  pClCtx->coopLevel = CoopLevel_ReadOnly;
  for (i = 0; i < ADMXRC_NUM_SPACE_MAX; i++) {
    pClCtx->spaceInfo[i].virtualBase = NULL;
    pClCtx->spaceInfo[i].virtualSize = 0;
  }
  os_set_client_context(pClObj, pClCtx);

 done:
  return status;
}

static OS_STATUS
onClose(
  OS_CLIENT_OBJECT* pClObj)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_DEVICE_OBJECT* pDevObj;
  DeviceContext* pDevCtx;
  ClientContext* pClCtx;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(1, ("onClose: entered, client = 0x%p\n", pClObj));

  pDevObj = os_get_device_object(pClObj);
  pDevCtx = os_get_device_context(pDevObj);
  pClCtx = os_get_client_context(pClObj);

  os_acquire_spinlock(&pDevCtx->owner.lock, &f);
  if (pClObj == pDevCtx->owner.exclusive) {
    os_release_spinlock(&pDevCtx->owner.lock, f);
    ADMXRC_CleanupDma(pDevCtx);
    ADMXRC_CleanupDMATargets(pDevCtx);
    ADMXRC_UnregAllInterruptEvents(pDevCtx);
    os_acquire_spinlock(&pDevCtx->owner.lock, &f);
    pDevCtx->owner.exclusive = NULL;
    os_release_spinlock(&pDevCtx->owner.lock, f);
  } else {
    os_release_spinlock(&pDevCtx->owner.lock, f);
  }

  if (pClCtx->coopLevel == CoopLevel_Shared) {
    os_acquire_spinlock(&pDevCtx->owner.lock, &f);
    pDevCtx->owner.numShared--;
    os_release_spinlock(&pDevCtx->owner.lock, f);
  }

  os_kernel_free(pClCtx);
  os_set_client_context(pClObj, NULL);

  return status;
}

static OS_STATUS
onMap(
  OS_CLIENT_OBJECT* pClObj,
  OS_MAPPABLE_REGION* region,
  void* context,
  void* virtual,
  os_uint64 phys,
  os_uint64 length)
{
  DeviceContext* pDevCtx;
  ClientContext* pClCtx;
  OS_DEVICE_OBJECT* pDevObj;
  unsigned int index = (unsigned int) (unsigned long) context;

  os_dbgprint(2, ("onMap: entered, context = 0x%p virtual = 0x%p phys=0x%8.8lx%8.8lx length=0x%8.8lx%8.8lx\n",
		  context,
		  virtual,
		  (unsigned long) os_uint64_hi(phys), (unsigned long) os_uint64_lo(phys),
		  (unsigned long) os_uint64_hi(length), (unsigned long) os_uint64_lo(length)));

  pDevObj = os_get_device_object(pClObj);
  pDevCtx = os_get_device_context(pDevObj);
  pClCtx = os_get_client_context(pClObj);

  pClCtx->spaceInfo[index].virtualBase = virtual;
  pClCtx->spaceInfo[index].virtualSize = (unsigned long) length;

  return OS_STATUS_SUCCESS;
}

static OS_STATUS
onIoctl(
  OS_CLIENT_OBJECT* pClObj,
  OS_IO_REQUEST*    req)
{
  OS_DEVICE_OBJECT* pDevObj;
  DeviceContext* pDevCtx;
  ClientContext* pClCtx;
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_SPINLOCK_FLAGS f;
  os_boolean bHandled;
  os_boolean bReadOnly, bExclusive;

  os_dbgprint(4, ("onIoctl: entered, client=0x%p code=0x%lx(%ld) inSize=0x%lx outSize=0x%lx\n",
		  pClObj,
		  (unsigned long) req->param.code,
		  (unsigned long) req->param.code,
		  (unsigned long) req->param.inSize,
		  (unsigned long) req->param.outSize));

  pDevObj = os_get_device_object(pClObj);
  pDevCtx = os_get_device_context(pDevObj);
  pClCtx = os_get_client_context(pClObj);

  os_acquire_spinlock(&pDevCtx->owner.lock, &f);
  bReadOnly = (pClCtx->coopLevel == CoopLevel_ReadOnly);
  bExclusive = (pClObj == pDevCtx->owner.exclusive);
  os_release_spinlock(&pDevCtx->owner.lock, f);

  bHandled = TRUE; /* Assume it will be handled in the next switch statement */
  switch (req->param.code) {
  case ADMXRC_IOCTL_OPENCARD2:
    status = ADMXRC_OpenCard2(pDevCtx, pClObj, req);
    break;

  case ADMXRC_IOCTL_GETBANKINFO:
    status = ADMXRC_GetBankInfo(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_GETCARDINFO2:
    status = ADMXRC_GetCardInfo2(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_GETFLASHINFO:
    status = ADMXRC_GetFlashInfo(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_GETFLASHBLOCKINFO:
    status = ADMXRC_GetFlashBlockInfo(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_GETSPACEINFO:
    status = ADMXRC_GetSpaceInfo(pDevCtx, pClCtx, req);
    break;

  case ADMXRC_IOCTL_GETSPACECONFIG:
    status = ADMXRC_GetSpaceConfig(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_GETVERSIONINFO:
    status = ADMXRC_GetVersionInfo(pDevCtx, req);
    break;

  default:
    /*
    ** The IOCTL wasnt handled. Try the switch statement below,
    ** if the calling thread owns the card.
    */
    bHandled = FALSE;
    break;
  }
  
  if (bHandled) {
    goto finish;
  }

  if (bReadOnly) {
    req->status.status = OS_STATUS_CANNOTSHARE;
    os_complete_request(req);
    return status;
  }

  switch (req->param.code) {
  /* High priority IOCTLs */
  case ADMXRC_IOCTL_DODMA:
    status = ADMXRC_DoDMA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_DODMA2D:
    status = ADMXRC_DoDMA2D(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_DODMAIMMEDIATE:
    status = ADMXRC_DoDMAImmediate(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_DODMAIMMEDIATE2D:
    status = ADMXRC_DoDMAImmediate2D(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_WAITFORINTERRUPT:
    status = ADMXRC_WaitForInterrupt(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_READ:
    status = ADMXRC_Read(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_WRITE:
    status = ADMXRC_Write(pDevCtx, req);
    break;

  /* Low priority IOCTLs */
  case ADMXRC_IOCTL_PROGRAMBYTES:
    status = ADMXRC_ProgramBytes(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_READREG:
    status = ADMXRC_ReadReg(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_WRITEREG:
    status = ADMXRC_WriteReg(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_ERASEFLASH:
    status = ADMXRC_EraseFlash(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_CANCELWAITFORINTERRUPT:
    status = ADMXRC_CancelWaitForInterrupt(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_REGINTERRUPTEVENT:
    status = ADMXRC_RegInterruptEvent(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_UNREGINTERRUPTEVENT:
    status = ADMXRC_UnregInterruptEvent(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_LOADFPGA:
    status = ADMXRC_LoadFPGA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_LOADFPGADMA:
    status = ADMXRC_LoadFPGADMA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_LOCALBUSRESET:
    status = ADMXRC_LocalBusReset(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_MAPDIRECTMASTER:
    status = ADMXRC_MapDirectMaster(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_READCONFIG:
    status = ADMXRC_ReadConfig(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_READFLASH:
    status = ADMXRC_ReadFlash(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_SETCLOCK2:
    status = ADMXRC_SetClock2(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_SETSPACECONFIG:
    status = ADMXRC_SetSpaceConfig(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_SETUPDMA:
    status = ADMXRC_SetupDMA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_SETUPDMAREMOTE:
    status = ADMXRC_SetupDMARemote(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_SYNCDIRECTMASTER:
    status = ADMXRC_SyncDirectMaster(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_UNSETUPDMA:
    status = ADMXRC_UnsetupDMA(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_WRITECONFIG:
    status = ADMXRC_WriteConfig(pDevCtx, req);
    break;

  case ADMXRC_IOCTL_WRITEFLASH:
    status = ADMXRC_WriteFlash(pDevCtx, req);
    break;

  default:
    req->status.status = OS_STATUS_INVALIDIOCTL;
    os_complete_request(req);
    return status;
  }

 finish:
  if (!OS_PENDING(status)) {
    req->status.status = status;
    os_complete_request(req);
  }

  return status;
}

/*
** -----------------------------------------------------------------------------
** The one and only driver parameters structure
** -----------------------------------------------------------------------------
*/

DriverParams admxrcDriverParams;

/*
** -----------------------------------------------------------------------------
** Required by framework
** -----------------------------------------------------------------------------
*/

const char* driver_name = DRIVER_NAME;

OS_STATUS
admxrc_driver_init(
  OS_DRIVER_OBJECT* pDrvObj)
{
  int i = 0;

  os_dbgprint(1, ("admxrc_driver_init: entered\n"));

  os_assert(sizeof(CONTROLXRC) == 0x10);
  os_assert(sizeof(PLX9080REGS) == 0x100);
  os_assert(sizeof(PLX9656REGS) == 0x200);
  os_assert(sizeof(DMACON9080) == 0x14);
  os_assert(sizeof(DMACON9656) == 0x14);
  os_assert(sizeof(XPLDMACON) == 0x20);
  os_assert(sizeof(XPLBRIDGEREGS) == 0x100);
  os_assert(sizeof(XPLSPACE1REGS) == 0x1000);

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9080;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9080;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC2L;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC2;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXPL;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXPL_PCIX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXPL;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXPL_PCI;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADPWRC2;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADPDRC2;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXP;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXP_PCIX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXP;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXP_PCI;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADPXPI;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADPXPI_PCI;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADPXPI;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADPXPI_PCIX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC4LS;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC4LX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_PLX;
  devDescs[i].id.pci.device = PCI_DEVICE_PLX_PCI9656;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC4SX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXRC4FX;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC4FX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADPEXRC4FX;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADPEXRC4FX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXRC5LX;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC5LX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXRC5T1;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC5T1;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADMXRC5LXA;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADMXRC5LXA;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_PCI;
  devDescs[i].id.pci.vendor = PCI_VENDOR_ADATA;
  devDescs[i].id.pci.device = PCI_DEVICE_ADATA_ADCPXRC4LX;
  devDescs[i].id.pci.subVendor = PCI_SUBVENDOR_ADATA;
  devDescs[i].id.pci.subDevice = PCI_SUBDEVICE_ADATA_ADCPXRC4LX;
  devDescs[i].id.pci.match =
    OS_DEVICE_DESC_MATCH_PCI_VENDOR |
    OS_DEVICE_DESC_MATCH_PCI_DEVICE |
    OS_DEVICE_DESC_MATCH_PCI_SUBVENDOR |
    OS_DEVICE_DESC_MATCH_PCI_SUBDEVICE;
  i++;

  devDescs[i].type = OS_BUSTYPE_INVALID;

  pDrvObj->hardwareIds = devDescs;

  pDrvObj->drvName = DRIVER_NAME;

  pDrvObj->onCreateDevice = onCreateDevice;
  pDrvObj->onStartDevice = onStartDevice;
  pDrvObj->onStopDevice = onStopDevice;
  pDrvObj->onDeleteDevice = onDeleteDevice;
  pDrvObj->onOpen = onOpen;
  pDrvObj->onClose = onClose;
  pDrvObj->onMap = onMap;
  pDrvObj->onIoctl = onIoctl;

  admxrcInitDriverParams(pDrvObj);

  return 0;
}

void
admxrc_driver_uninit(
  OS_DRIVER_OBJECT* pDrvObj)
{
  os_dbgprint(1, ("admxrc_driver_uninit: entered\n"));
}
