/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods_dma.c - Model-specific routines for DMA transfers.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

static void
dumpDescriptorsPlxbased(
  DeviceContext* pDevCtx,
  PLXDESC* desc,
  unsigned int channel)
{
  unsigned int i = 0;
  os_uint32    offsetl;
  os_uint32    pci, local, count, link;  

  os_dbgprint(0, ("Dump of PLXDESCs:\n"));
  os_dbgprint(0, ("###### DESC------ DESC-BUS---------- [PCI-------- LOCAL----- COUNT----- LINK------]\n"));
	
  offsetl = sizeof(PLXDESC);
  while (1) {
    pci = os_le32_to_cpu(desc->pci);
    local = os_le32_to_cpu(desc->local);
    count = os_le32_to_cpu(desc->count);
    link = os_le32_to_cpu(desc->link);
 
    os_dbgprint(0, ("%6.6ld 0x%p 0x%8lx [%8.8lx %8.8lx %8.8lx %8.8lx]\n",
		   (unsigned long) i,
		   (void*) desc,
		   (unsigned long) pDevCtx->dma.channel[channel].descriptors.pci[i],
		   (unsigned long) pci,
		   (unsigned long) local,
		   (unsigned long) count,
		   (unsigned long) link));
    if (link & PLXDESCF_ENDCHAIN) {
      break;
    }
    desc++;
    i++;
  }
}

static void
dumpDescriptorsXplbased(
  DeviceContext* pDevCtx,
  XPLDMADESC* desc,
  unsigned int channel)
{
  unsigned int i = 0;
  os_uint32 offsetl;
  os_uint32 pah, pal, la, siz, ndh, ndl;
	
  os_dbgprint(0, ("Dump of DMA linked list:\n"));
  os_dbgprint(0, ("### DESC-VIRT- DESC-BUS-- { PCI--------------- LOCAL----- COUNT----- LINK-------------- }\n"));
	
  offsetl = sizeof(XPLDMADESC);
  while (1) {
    pal = os_le32_to_cpu(desc->PAL);
    pah = os_le32_to_cpu(desc->PAH);
    la = os_le32_to_cpu(desc->LA);
    siz = os_le32_to_cpu(desc->SIZ);
    ndl = os_le32_to_cpu(desc->NDL);
    ndh = os_le32_to_cpu(desc->NDH);
    os_dbgprint(0, ("%3.3ld 0x%p 0x%8.8lx { 0x%8.8lx%8.8lx 0x%8.8lx 0x%8.8lx 0x%8.8lx%8.8lx }\n",
		    (unsigned long) i,
		    (void*) desc,
		    (unsigned long) pDevCtx->dma.channel[channel].descriptors.pci[i],
		    (unsigned long) pah, 
		    (unsigned long) pal,
		    (unsigned long) la,
		    (unsigned long) siz,
		    (unsigned long) ndh,
		    (unsigned long) ndl));
    if (ndl & XPLDMADESC_ENDCHAIN) {
      break;
    }
    desc++;
    i++;
  }
}

/*
** -----------------------------------------------------------------------
** MethodAddDmaBuffer
** -----------------------------------------------------------------------
*/

MethodResult
methodAddDmaBufferPlxbased(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  unsigned int    index,
  DmaScatterNode* snode)
{
  PLXDESC*  desc;
  os_uint32 flags;

  /* Set up PLX descriptor */
  desc = &pDevCtx->dma.channel[channel].descriptors.aligned.plxbased[index];
  if (index) {
    (desc - 1)->link |= os_cpu_to_le32(pDevCtx->dma.channel[channel].descriptors.pci[index]);
  }
  desc->pci   = os_cpu_to_le32(snode->busAddr);
  desc->local = os_cpu_to_le32(snode->localAddr);
  desc->count = os_cpu_to_le32(snode->length);
  flags = snode->bWrite ? 0x1U : 0x9U;
  desc->link  = os_cpu_to_le32(flags);

  return MethodResult_Success;
}

MethodResult
methodAddDmaBufferXplrev0(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  unsigned int    index,
  DmaScatterNode* snode)
{
  return MethodResult_NotSupported;
}

MethodResult methodAddDmaBufferXplrev1(
  DeviceContext* pDevCtx,
  unsigned int channel,
  unsigned int index,
  DmaScatterNode* scatterNode)
{
  XPLDMADESC* desc;
  os_uint32 sizFlags;

  desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[index];
  if (index) {
    (desc - 1)->NDL |= os_cpu_to_le32(pDevCtx->dma.channel[channel].descriptors.pci[index]);
  }
  if (scatterNode->bWrite) {
    sizFlags = XPLDMADESC_PCITOLOCAL;
  } else {
    sizFlags = XPLDMADESC_LOCALTOPCI;
  }
  desc->PAL = os_cpu_to_le32((os_uint32) scatterNode->busAddr);
  desc->PAH = os_cpu_to_le32(0);
  desc->LA = os_cpu_to_le32(scatterNode->localAddr);
  desc->SIZ = os_cpu_to_le32(scatterNode->length | sizFlags);
  desc->NDL = os_cpu_to_le32(XPLDMADESC_PCISPACE);
  desc->NDH = os_cpu_to_le32(0);
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** MethodDoDma
** -----------------------------------------------------------------------
*/

static MethodResult
methodDoDmaPlxbased(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  os_uint16       shadow_dmacsr,
  unsigned int    numBuffer)
{
  MethodResult res = MethodResult_Success;
  os_uint16    mask, val16;
  PLXDESC*     desc;

  os_dbgprint(6, ("methodDoDmaPlxbased: channel=%ld start=%ld mode=0x%lx, shadow.dmacsr=0x%4.4lx\n",
		  (unsigned long) channel,
		  (unsigned long) start,
		  (unsigned long) mode,
		  (unsigned long) shadow_dmacsr));
      
  if (start) {
    /*
    ** Kick off a chunk of a DMA transfer
    */

    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.plxbased[numBuffer - 1];
    desc->link |= os_cpu_to_le32(PLXDESCF_ENDCHAIN | PLXDESCF_INTERRUPT);

    if (os_dbglevel(8)) {
      dumpDescriptorsPlxbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.plxbased,
			      channel);
    }
  
    mode |=
      (0x1U <<  9)  | /* Use chained mode           */
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    mask = channel ? 0x0200U : 0x0002U;
    val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
    if (val16 & (mask << 3)) {
      /* Set current transfer mode (chained mode bit added) */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].MODE, mode);
      /* Set current transfer size to 0 */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].SIZE, 0);
      /* Set the descriptor pointer */
      os_write_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].DPR,
			     pDevCtx->dma.channel[channel].descriptors.pci[0] | 0x1U);

      /* Make sure the writes have completed */
      os_read_register_le32(&pDevCtx->hw.adapter.pci9080->DMA[channel].DPR);
      
      /* Strobe the start bit */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, shadow_dmacsr | mask);
    } else {
      os_dbgprint(0, ("*** methodDoDmaPlxbased: DMA channel %ld not idle\n", (unsigned long) channel));
      res = MethodResult_GeneralFailure;
    }
  } else {
    /*
    ** Abort a chunk of a DMA transfer
    */

    mask = channel ? 0x0100 : 0x0001;
  
    /* First, disable the DMA channel */
    os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, shadow_dmacsr & ~mask);
    
    /* Now check if DMA channel is active or not */
    val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
    if (val16 & (mask << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaPlxbased: nothing to abort on DMA channel %ld\n", (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, shadow_dmacsr | (mask << 2));
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(50);
      /* Did we succeed in aborting it? */
      val16 = os_read_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord);
      if (val16 & (mask << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(3, ("methodDoDmaPlxbased: aborted DMA on channel %ld\n", (unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaPlxbased: failed to abort DMA on channel %ld!\n", (unsigned long) channel));
	res = MethodResult_GeneralFailure;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that will have
    ** occurred if we aborted.
    */
    os_write_register_le16(&pDevCtx->hw.adapter.pci9080->DMACSR.asWord, shadow_dmacsr | mask | (mask << 3));
  }

  return res;
}

MethodResult
methodDoDmaXrc(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  return methodDoDmaPlxbased(pDevCtx, channel, start, mode, pDevCtx->shadow.regs.plxbased.dmacsr, numBuffer);
}

MethodResult
methodDoDmaXplrev0(
  DeviceContext*  pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  return MethodResult_NotSupported;
}

MethodResult
methodDoDmaXplrev1(
  DeviceContext* pDevCtx,
  unsigned int    channel,
  os_boolean      start,
  os_uint32       mode,
  unsigned int    numBuffer)
{
  XPLBRIDGEREGS* regs;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* desc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;

  regs = pDevCtx->hw.adapter.xpl;
		
  if (start) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[numBuffer - 1];
    desc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    descBus = pDevCtx->dma.channel[channel].descriptors.pci[0];
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
	/* Set current transfer mode (adds in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode | (0x1U << 9));
	/* Set the descriptor pointer & flags */
	os_write_register_le32(&regs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
	os_write_register_le32(&regs->DMA[channel].NDH, 0);
      } else {
	desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[0];
	/* Set current transfer mode (DOESN'T add in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode);
	/* Set current transfer PCI address */
	os_write_register_32(&regs->DMA[channel].PAL, desc->PAL);
	os_write_register_32(&regs->DMA[channel].PAH, desc->PAH);
	/* Set current transfer local address */
	os_write_register_32(&regs->DMA[channel].LA, desc->LA);
	/* Set current transfer size */
	os_write_register_32(&regs->DMA[channel].SIZ, desc->SIZ);
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    if (0) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(&regs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that
    ** will have occurred if we aborted.
    */
    dxctl |= 0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

MethodResult
methodDoDmaXrc4ls(
  DeviceContext*  pDevCtx,
  unsigned int channel,
  os_boolean start,
  os_uint32 mode,
  unsigned int numBuffer)
{
  return methodDoDmaPlxbased(pDevCtx, channel, start, mode, pDevCtx->shadow.regs.xrc4ls.dmacsr, numBuffer);
}

MethodResult
methodDoDmaXrce4fx(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean start,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XPLBRIDGEREGS* regs;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* desc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;
  
  regs = pDevCtx->hw.adapter.xpl;
  
  if (start) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[numBuffer - 1];
    desc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    descBus = pDevCtx->dma.channel[channel].descriptors.pci[0];
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
	/* Set current transfer mode (adds in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode | (0x1U << 9));
	/* Set the descriptor pointer & flags */
	os_write_register_le32(&regs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
	os_write_register_le32(&regs->DMA[channel].NDH, 0);
      } else {
	desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[0];
	/* Set current transfer mode (DOESN'T add in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode);
	/* Set current transfer PCI address */
	os_write_register_32(&regs->DMA[channel].PAL, desc->PAL);
	os_write_register_32(&regs->DMA[channel].PAH, desc->PAH);
	/* Set current transfer local address */
	os_write_register_32(&regs->DMA[channel].LA, desc->LA);
	/* Set current transfer size */
	os_write_register_32(&regs->DMA[channel].SIZ, desc->SIZ);
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    if (0) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(&regs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that
    ** will have occurred if we aborted.
    */
    dxctl |= 0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}

MethodResult
methodDoDmaXrc4fx(
  DeviceContext* pDevCtx,
  unsigned int channel,
  os_boolean start,
  os_uint32 mode,
  unsigned int numBuffer)
{
  XPLBRIDGEREGS* regs;
  MethodResult result = MethodResult_Success;
  XPLDMADESC* desc;
  os_uint32 dxctl, val32;
  os_uint32 descBus;
  
  regs = pDevCtx->hw.adapter.xpl;
  
  if (start) {
    /* Fix up last descriptor to end the chain and generate an interrupt */
    desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[numBuffer - 1];
    desc->NDL |= os_cpu_to_le32(XPLDMADESC_ENDCHAIN);
    
    mode |=
      (0x1U << 10)  | /* "DONE" interrupt enable    */
      (0x1U << 17);   /* route interrupt to PCI bus */
    if (admxrcDriverParams.pciWriteAndInvalidate) {
      mode |= (0x1U << 13); /* Enable use of PCI write-and-invalidate command */
    }

    if (os_dbglevel(8)) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    descBus = pDevCtx->dma.channel[channel].descriptors.pci[0];
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 5)) {
      if (numBuffer > 1) {
	/* Set current transfer mode (adds in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode | (0x1U << 9));
	/* Set the descriptor pointer & flags */
	os_write_register_le32(&regs->DMA[channel].NDL, descBus | XPLDMADESC_PCISPACE);
	os_write_register_le32(&regs->DMA[channel].NDH, 0);
      } else {
	desc = &pDevCtx->dma.channel[channel].descriptors.aligned.xpl[0];
	/* Set current transfer mode (DOESN'T add in chained-mode bit) */
	os_write_register_le32(&regs->DMA[channel].CFG, mode);
	/* Set current transfer PCI address */
	os_write_register_32(&regs->DMA[channel].PAL, desc->PAL);
	os_write_register_32(&regs->DMA[channel].PAH, desc->PAH);
	/* Set current transfer local address */
	os_write_register_32(&regs->DMA[channel].LA, desc->LA);
	/* Set current transfer size */
	os_write_register_32(&regs->DMA[channel].SIZ, desc->SIZ);
      }

      /* Strobe the start bit */
      dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x1U);
    } else {
      os_dbgprint(0, ("*** methodDoDmaXplrev1: DMA not idle, channel=%ld\n",
		      (unsigned long) channel));
    }
  } else {
    os_dbgprint(2, ("methodDoDmaXplrev1: aborting, channel=%ld\n",
		    (unsigned long) channel));
    
    if (0) {
      dumpDescriptorsXplbased(pDevCtx,
			      pDevCtx->dma.channel[channel].descriptors.aligned.xpl,
			      channel);
    }
    
    /* First, disable the DMA channel */
    dxctl = pDevCtx->shadow.regs.xpl.dxctl[channel];
    dxctl &= ~0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
    /* Wait a bit for the bridge to abort the DMA */
    os_delay_busy_us(10);
    /* Now check if DMA channel is active or not */
    val32 = os_read_register_le32(&regs->DMA[channel].CTL);
    if (val32 & (0x1U << 4)) {
      /* DMA is not active so must have already finished */
      os_dbgprint(1, ("methodDoDmaXplrev1: nothing to do, channel=%ld\n",
		      (unsigned long) channel));
    } else {
      /* DMA is active - abort it */
      os_write_register_le32(&regs->DMA[channel].CTL, dxctl | 0x2U);
      /* Wait a bit for the PCI9080/PCI9656 to abort the DMA */
      os_delay_busy_us(10);
      /* Did we succeed in aborting it? */
      val32 = os_read_register_le32(&regs->DMA[channel].CTL);
      if (val32 & (0x1U << 4)) {
	/* Yes, was aborted ok */
	os_dbgprint(1, ("methodDoDmaXplrev1: aborted, channel=%ld\n",
			(unsigned long) channel));
      } else {
	/* No! Not much we can do now, except restore things */
	os_dbgprint(0, ("*** methodDoDmaXplrev1: failed to abort, channel=%ld!\n",
			(unsigned long) channel));
	result = MethodResult_Timeout;
      }
    }
    
    /*
    ** Reenable the DMA channel and clear the DMA interrupt that
    ** will have occurred if we aborted.
    */
    dxctl |= 0x4U;
    os_write_register_le32(&regs->DMA[channel].CTL, dxctl);
  }
  
  return result;
}
